# Copyright (C) 2009  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


package Xyne::Arch::PkgD;
use strict;
use warnings;

use threads;
use Thread::Queue;
use IO::Socket;
use Xyne::Common::Log;

our $VERSION = '0.15';

my $OPTIONS = {
  'Cache'    => 'str',
  'Config'   => 'str',
  'Interval' => 'int',
  'Log'      => 'str',
  'Master'   => 'str',
  'Port'     => 'int',
  'Quiet'    => 'bool'
};

my $OPTION_MAP = {
  '--cache'    => 'Cache',
  '--config'   => 'Config',
  '--get-pid'  => 'action',
  '--help'     => 'action',
  '--interval' => 'Interval',
  '--log'      => 'Log',
  '--master'   => 'Master',
  '--port'     => 'Port',
  '--quiet'    => 'Quiet'
};

my $DEFAULTS = {
  'Cache'    => ['/var/cache/pacman/pkg'],
  'Config'   => '/etc/pkgd.conf',
  'Interval' => 300,
  'Log'      => undef,
#  'Master'   => [],
  'MaxDepth' => 3,
  'Port'     => 14687,
  'Quiet'    => 0
};

my @LISTS = qw/Cache Master Slave/;




sub new {
  my $proto = shift;
  my $class = ref($proto) || $proto;
  my $self  = {};
  bless ($self, $class);
  return $self;
}




sub display_help
{
  my ($self) = @_;
  print <<HELP;
USAGE
  pkgd [options]

OPTIONS
  '--cache'    specify cache*
  '--config'   specify config
  '--get-pid'  get pid**
  '--help'     display this message
  '--interval' set notification interval
  '--log'      set log
  '--master'   specify master pkgd*
  '--port'     set port
  '--quiet'    suppress output

* multiple may be specified
** this will return the pid of the pkgd running on the port set by /etc/pkgd.conf

See the man page for more information.
HELP
  exit;
}

sub get_arg
{
  my ($self,$arg) = @_;
  return if not defined($arg);
  my $value = $self->{ARGS}->{$arg};
  if (grep {$_ eq $arg} @LISTS)
  {
    return () if not defined($value);
    return @{ $value };
  }
  return $value;
}

sub check_for_file
{
  my ($self,$file) = @_;
  return defined($self->get_file_path($file));
}

sub clear_option
{
  my ($self,$option) = @_;
  return if not defined($option);
  delete $self->{OPTIONS}->{$option};
}

sub error
{
  my ($self,$msg) = @_;
  print $msg;
  exit 1;
}

sub get_file_path
{
  my ($self,$file) = @_;
  foreach my $cache ($self->get_option('Cache'))
  {
    my $path = $cache.'/'.$file;
    if (-e $path and -f $path)
    {
      return $path;
    }
  }
  return undef;
}

sub get_file_query_response
{
  my ($self,$args) = @_;
  my ($found,$checked) = $self->process_file_query($args);
  my $response = "PKGD\n[FileQueryResults]\n";
  foreach my $file (keys %{ $found })
  {
    my $server = $found->{$file};
    $response .= "File:${file}::Server:$server\n";
  };
  foreach my $server (keys %{ $checked })
  {
    $response .= "Checked:$server\n";
  };
  return $response;
}

sub get_option
{
  my ($self,$option) = @_;
  return if not defined($option);
  my $value = $self->{OPTIONS}->{$option};
  $value = $DEFAULTS->{$option} if not defined($value);
  if (grep {$_ eq $option} @LISTS)
  {
    return () if not defined($value);
    return @{ $value };
  }
  return $value;
}

sub get_pid
{
  my ($self) = @_;
  my $sock = new IO::Socket::INET(
    PeerAddr => '127.0.0.1',
    PeerPort => $self->get_option('Port'),
    Proto => 'tcp')
  #|| die "Error creating socket: $!";
  || return "";

  print $sock "PKGD\n[GetPID]\r\n\r\n";
  $sock->flush;
  my $response = &read_socket($sock);
  return $response if $response =~ m/^\d+$/;
}

sub handle_file_request
{
  my ($self,$client,$repo,$file,$request) = @_;
  my $client_ip = $client->peerhost;
  my ($found,$checked) = $self->search_for_files($file);
  if (defined($found->{$file}))
  {
    if ($found->{$file} eq '%SELF%')
    {
      my $path = $self->get_file_path($file);
      $self->return_file($client,$path,$request);
    }
    else
    {
      print $client "HTTP/1.1 303 See Other\r\n";
      print $client "Location: http://$found->{$file}/$repo/$file\r\n\r\n";
      close $client;
    }
  }
  else
  {
    print $client "HTTP/1.1 404 Not Found\r\n\r\n";
    close $client;
  }
}

sub handle_request
{
  my ($self,$client) = @_;
  my $client_ip = $client->peerhost;
  my $request = &read_socket($client);
  if ($request =~ m/^PKGD/)
  {
    my $type = '';
    my %files;
    my %checked;
    my $depth;
    my $max_depth;
    foreach my $line (split /\s*\n\s*/,$request)
    {
      # match [*] to get the type
      if ($line =~ m/^\[([^\]]+)\]$/)
      {
        $type = $1;
        if ($type eq 'GetPID')
        {
          $self->message("pid request from $client_ip\n");
          print $client "$$\r\n\r\n" if ($client_ip eq '127.0.0.1');
          close $client;
          return;
        }
      }
      elsif ($type eq 'Slave' and $line =~ m/^Port:(\d+)(?:::NetworkID:(.+))?$/)
      {
        if (not defined($2) or (defined($2) and $2 eq $self->get_option('NetworkID')))
        {
          $self->set_option('Slave',"$client_ip:$1");
          $self->message("accepted slave $client_ip:$1\n");
        }
        else
        {
          $self->message("rejected slave $client_ip:$1\n");
        }
        close $client;
        return;
      }
      elsif ($type eq 'FileQuery')
      {
        if ($line =~ m/^File:(\S+)$/)
        {
          $files{$1} = 0;
        }
        elsif ($line =~ m/^Checked:(\S+)$/)
        {
          $checked{$1} = 1;
        }
        elsif ($line =~ m/^Depth:(\d+)$/)
        {
          $depth = $1;
        }
        elsif ($line =~ m/^MaxDepth:(\d+)$/)
        {
          $max_depth = $1;
        }
      }
    }
    if ($type eq 'FileQuery')
    {
      $depth = 0 if not defined($depth);
      my $args = {
        'checked'   => \%checked,
        'depth'     => $depth,
        'files'     => \%files,
        'max_depth' => $max_depth
      };
      my $response = $self->get_file_query_response($args);
      print $client $response;
    }
    print $client "\r\n\r\n";
    close $client;
  }
  elsif ($request =~ m/^GET \/([^\/]+)\/(.+?) HTTP\/\d\.\d\r$/m)
  {
    my ($repo,$file) = ($1,$2);
    $self->message("client $client_ip requested $file\n");
    # For testing threads vs forks
    if (0)
    {
      my $thr = threads->create('handle_file_request',$self,$client,$repo,$file,$request);
      $thr->detach();
    }
    else
    {
      my $pid = fork();
      if ($pid == 0)
      {
        $self->handle_file_request($client,$repo,$file,$request);
      }
    }
  }
  else
  {
    $self->message("dropped request from $client_ip:\n$request\n");
    $request =~ s/\r/\\r/g;
    print $request;
    close $client;
  }
}

sub listen
{
  my ($self) = @_;
  my $port = $self->get_option('Port');
  my $sock = new IO::Socket::INET(
    LocalPort => $port,
    Listen => 1,
    Reuse => 1,
    Proto => 'tcp'
  ) || $self->error("Error creating socket: $!\n");
  $self->message("listening on port $port\n");

  while(my $client = $sock->accept())
  {
    $self->handle_request($client);
  }
  close($sock);
}

sub load_conf
{
  my ($self) = @_;
  my $path = $self->get_option('Config');
  open (my $fh, '<', $path) or $self->error("Unable to open $path: $!\n");
  while (defined(my $line=<$fh>))
  {
    next if ($line =~ m/^\s*#/ or $line =~ m/^\s*$/);
    if ($line =~ m/^\s*(\S+)\s*=\s*(.*?)\s*$/)
    {
      my ($option,$value) = ($1,$2);
      $self->set_option($option,$value) if defined($OPTIONS->{$option});
    }
    elsif ($line =~ m/^\s*(\S+)\s*$/)
    {
      my $option = $1;
      $self->set_option($option,1) if defined($OPTIONS->{$option});
    }
  }
  close $fh;
}

sub maybe_notify_masters
{
  my ($self) = @_;
  my @masters = $self->get_option('Master');
  my %masters;
  foreach my $master (@masters)
  {
    my ($url,$port) = &get_url_and_port($master);
    $masters{$url} = $port;
  }
  if (keys(%masters))
  {
    #my $thr = $self->{'NOTIFICATION_THREAD'};
    #$thr->kill('KILL')->detach() if defined($thr);
    my $args;
    $args->{'Parent'} = $$;
    $args->{'Port'} = $self->get_option('Port');
    $args->{'NetworkID'} = $self->get_option('NetworkID');
    $args->{'Masters'} = \%masters;
    my $thr = threads->create('notify_masters',$args);
    $self->{'NOTIFICATION_THREAD'} = $thr;
  }
}

sub message
{
  my ($self,$msg) = @_;
  if (not $self->get_option('Quiet'))
  {
    print $msg;
  }
  if (defined($self->{LOG}) or $self->get_option('Log'))
  {
    $self->{LOG} = Xyne::Common::Log->new( $self->get_option('Log') ) if not defined($self->{LOG});
    $self->{LOG}->log($msg);
  }
}

sub parse_args
{
  my ($self,@args) = @_;
  
  while (@args)
  {
    my $arg = shift @args;
    next if not defined($OPTION_MAP->{$arg});
    if ($arg eq '--get-pid')
    {
      print $self->get_pid;
      exit;
    }
    elsif ($arg eq '--help')
    {
      print $self->display_help;
      exit;
    }
    elsif ($arg eq '--quiet')
    {
      $self->set_option_by_arg('Quiet',1);
    }
    else
    {
      my $parameter = @args ? shift @args : '';
      $self->set_option_by_arg($OPTION_MAP->{$arg},$parameter);
    }
  }
}

sub process_file_query
{
  my ($self,$args) = @_;
  my %files = %{ $args->{'files'} };
  my %checked = %{ $args->{'checked'} };
  my $depth = $args->{'depth'};
  my $max_depth = $args->{'max_depth'};
  
  my %found;
  foreach my $file (keys %files)
  {
    if ($self->check_for_file($file))
    {
      $files{$file}++;
      $found{$file} = '%SELF%';
    }
  }
  my @files = (grep {$files{$_} == 0} keys %files);
  my $too_deep = 0;
  if (defined($depth) and defined($max_depth) and $depth >= $max_depth)
  {
    $too_deep = 1;
  }
  if (@files and not $too_deep)
  {
    $depth++;
    foreach my $slave ($self->get_option('Slave'))
    {
      my %remaining_files;
      next if defined($checked{$slave}) and $checked{$slave} > 0;
      $checked{$slave} = 1;
      $remaining_files{$_} = 0 foreach @files;
      my ($slave_files,$slave_checked) = $self->query_slave_for_files($slave, {
        'checked'   => \%checked,
        'depth'     => $depth,
        'files'     => \%remaining_files,
        'max_depth' => $max_depth
      });
      $checked{$_} += $slave_checked->{$_} foreach keys %{ $slave_checked };
      my @files = ();
      foreach my $file (keys %{ $slave_files })
      {
        $found{$file} = $slave_files->{$file};
        $files{$file}++;
      }
      @files = (grep {$files{$_} == 0} keys %files);
    }
  }
  return (\%found,\%checked);
}

sub query_slave_for_files
{
  my ($self,$slave,$args) = @_;
  my $request = &format_file_query($args);
  my ($url,$port) = &get_url_and_port($slave);
  my $response = &send_request($request,$url,$port);
  $response =~ s/%SELF%/$slave/g;
  return &parse_file_query_results($response);
}

sub return_file
{
  #my $pid = fork();
  #return if $pid != 0;
  my ($self,$client,$file,$request) = @_;
  my $size = (stat($file))[7];
  my $range_start;
  my $range_end;
  if ($request =~ m/Range: bytes=(\d*)-(\d*)/)
  {
    $range_start = $1;
    $range_end = $2;
  }
  $range_start = 0 if not defined($range_start);
  $range_end = $size - 1 if not $range_end;
  my $remaining = $range_end - $range_start;
  my $header = "HTTP/1.1 200 OK\r\nAccept-Ranges: bytes\r\nContent-Length: $size\r\n";
  $header .= "Content-Range: bytes $range_start-$range_end/$size\r\n" if $remaining != $size;
  $header .= "Connection: close\r\nContent-Type: application/x-gzip\r\n\r\n";
  print $client $header;
  if (open(my $fh,'<',$file))
  {
    binmode $fh;
    binmode $client;
    seek($fh,$range_start,0) if $range_start > 0;
    my $data;
    my $step = 1024;
    while(defined($client) and my $r = read( $fh, $data, $step ))
    {
      print $client $data;
      $client->flush;
      $remaining -= $r;
      $step = $remaining + 1 if $remaining < $step;
    }
    close $fh;
  }
  #if ($range_end < $size or $remaining > 0)
  #{
  #  $self->handle_request($client);
  #}
  #else
  #{
    close($client);
  #}
  #exit() if $pid == 0;
}

sub run
{
  my ($self,@args) = @_;
  $self->load_conf();
  $self->parse_args(@args);
  $self->maybe_notify_masters;
  $self->listen;
}

sub search_for_files
{
  my ($self,@files) = @_;
  my %files;
  $files{$_} = 0 foreach @files;
  my %checked;
  my $depth = 0;
  my $max_depth = $self->get_option('MaxDepth');

  my $args = {
    'checked'   => \%checked,
    'depth'     => $depth,
    'files'     => \%files,
    'max_depth' => $max_depth
  };
  return $self->process_file_query($args);
}

sub set_option
{
  my ($self,$option,$value) = @_;
  return if not defined($option);
  if (grep {$_ eq $option} @LISTS)
  {
    if (not grep {$_ eq $option} @{ $self->{OPTIONS}->{$option} })
    {
      push @{ $self->{OPTIONS}->{$option} }, $value;
    }
  }
  else
  {
    $self->{OPTIONS}->{$option} = $value;
  }
}

sub set_option_by_arg
{
  my ($self,$option,$value) = @_;
  return if not defined($option);
  if ($option eq 'Config')
  {
    delete $self->{OPTIONS};
    $self->set_option($option,$value);
    $self->load_conf;
    foreach my $key (keys %{ $self->{ARGS} })
    {
      $self->set_option($key,$self->get_arg($key));
    }
  }
  else
  {
    delete $self->{OPTIONS}->{$option} if not $self->{ARGS}->{$option};
    if (grep {$_ eq $option} @LISTS)
    {
      if (not grep {$_ eq $option} @{ $self->{OPTIONS}->{$option} })
      {
        push @{ $self->{OPTIONS}->{$option} }, $value;
      }
    }
    else
    {
        $self->{ARGS}->{$option} = $value;
    }
    $self->set_option($option,$value);
  }
}




#####################################
########## CLASS FUNCTIONS ##########
#####################################

sub format_file_query
{
  my ($args) = @_;
  my %files = %{ $args->{'files'} };
  my %checked = %{ $args->{'checked'} };
  my $depth = $args->{'depth'};
  my $max_depth = $args->{'max_depth'};

  my $query = "PKGD\n[FileQuery]\n";
  $query .= "File:$_\n" foreach grep {$files{$_} == 0} keys %files;
  $query .= "Checked:$_\n" foreach grep {$_} keys %checked;
  $depth = 0 if not $depth;
  $query .= "Depth:$depth\n";
  $query .= "MaxDepth:$max_depth\n" if defined($max_depth);
  return $query;
}

sub get_url_and_port
{
  my ($server) = @_;
  if ($server =~ m/^(.*?):(\d+)$/)
  {
    return ($1,$2);
  }
  return ($server, $DEFAULTS->{'Port'});
}

sub notify_masters
{
  my ($args) = @_;
  my $parent_pid = $args->{'Parent'};
  my $port = $args->{'Port'};
  my $interval = $args->{'Interval'};
  $interval = $DEFAULTS->{'Interval'} if not defined($interval);
  my $network_id = $args->{'NetworkID'};
  my %masters = %{ $args->{'Masters'} };
  exit if not ($parent_pid and $port and %masters);
  sleep 3;
  while(kill 0,$parent_pid)
  {
    foreach my $ip (keys %masters)
    {
      my $m_port = $masters{$ip};
      if ( my $sock = new IO::Socket::INET(PeerAddr => $ip,PeerPort => $m_port,Proto => 'tcp') )
      {      
        print $sock "PKGD\n[Slave]\nPort:$port";
        print $sock "::NetworkID:$network_id" if defined($network_id);
        print $sock "\r\n\r\n";
        $sock->flush;
        close $sock;
      }
    }
    sleep $interval;
  }
}

sub parse_file_query_results
{
  my ($results) = @_;
  my $type = '';
  my %files;
  my %checked;
  foreach my $line (split /\s*\n\s*/,$results)
  {
    # match [*] to get the type
    if ($line =~ m/^\[([^\]]*)\]$/)
    {
      $type = $1;
    }
    elsif ($type eq 'FileQueryResults')
    {
      if ($line =~ m/^File:(\S+)::Server:(\S+)$/)
      {
        $files{$1} = $2;
      }
      elsif ($line =~ m/^Checked:(\S+)$/)
      {
        $checked{$1} = 1;
      }
    }
  }
  return (\%files,\%checked);
}

sub read_socket
{
  my ($sock) = @_;
  my $data = '';
  my $byte;
  while(read( $sock, $byte, 1 ))
  {
    $data .= $byte;
    last if (substr($data,-4) eq "\r\n\r\n");
  }
  $data = substr($data,0,-4);
  return $data;
}

sub send_request
{
  my ($request,$url,$port) = @_;
  my $sock = new IO::Socket::INET(
    PeerAddr => $url,
    PeerPort => $port,
    Proto => 'tcp')
  || return '';

  print $sock $request;
  print $sock "\r\n\r\n";
  $sock->flush;
  return &read_socket($sock); 
}

sub transfer_pkg
{
  my ($pkg,$address,$port,$cachedir,$queue) = @_;
  my $request = "GET /PkgD/$pkg HTTP/1.1\r\nUser-Agent: PkgD/$VERSION\r\n\r\n";
  
  my $sock = new IO::Socket::INET(
    PeerAddr => $address,
    PeerPort => $port,
    Proto => 'tcp')
  || return '';

  print $sock $request;
  $sock->flush;
  my $header = &read_socket($sock);
  
  if ($header =~ m/Content-Length: (\d+)/)
  {
    my $size = $1;
    my $data;
    my $step = ($size < 1024) ? $size : 1024;
    my $fpath = $cachedir.'/'.$pkg;
    open(my $fh, '>', $fpath);
    binmode $fh;
    while(my $l = read( $sock, $data, $step ))
    {
      print $fh $data;
      $size .= $l;
      $step = $size if $size < $step;
      last if $step <= 0;
    }
    close $fh;
    read( $sock, $data, 4);
  }
  $queue->enqueue($pkg);
}

sub retrieve_and_save_pkg_from_pkgd
{
  my @pkgs = @{$_[0]};
  my @pkgd_urls = @{$_[1]};
  my $cachedir = $_[2];
  my %pkgmap = ();
  foreach my $url (@pkgd_urls)
  {
    if ($url =~ m/^http:\/\/([^\/]*):(\d+)\/[^\/]+$/)
    {
      my $address = $1;
      my $port = $2;

      my $request = "PKGD\n[FileQuery]";
      foreach my $pkg (@pkgs)
      {
        $request .= "\nFile:$pkg";
      }
      $request .= "\r\n\r\n";
      
      my $response = &send_request($request,$address,$port);
      $response =~ s/%SELF%/${address}:${port}/g;
      foreach my $line (split(/\s*\n\s*/, $response))
      {
        if ($line =~ m/^File:(.+)::Server:(.+)$/)
        {
          $pkgmap{$1} = $2;
        }
      }
    }
  }
  my $queue = Thread::Queue->new();
  my @pending = ();
  foreach my $pkg (keys(%pkgmap))
  {
    if ($pkgmap{$pkg} =~ m/^([\d\.]+):(\d+)$/)
    {
      my ($address,$port) = ($1,$2);
      print "PkgD: requesting $pkg from $address:$port\n";
      my $thr = threads->create('transfer_pkg',$pkg,$address,$port,$cachedir,$queue);
      $thr->detach();
      push(@pending, $pkg);
    }
    else
    {
      print "PkgD: ignoring malformed address: " . $pkgmap{$pkg} . "\n";
    }
  }
  # I tried to "->join()" the threads, but they kept blocking indefinitely, even
  # with lines like
  # $thr->join() if $thr->is_joinable()
  # which, according to the documentation, should only be executed when the
  # thread can be joined without blocking.
  # Not true.
  foreach my $foo (@pending)
  {
    my $pkg = $queue->dequeue();
    print "PkgD: retrieved $pkg\n";
  }
}

1;

__END__

=head1 NAME

Xyne::Arch::PkgD - A class for sharing pacman packages over a LAN.

=head1 SYNOPSIS

  use Xyne::Arch::PkgD;

=head1 DESCRIPTION

A class to provide client and server functionality for sharing
pacman packages over a LAN to avoid unnecessary downloads from
the external network.

=head2 EXPORT

None by default.

=head1 AUTHOR

Xyne <lt>xyne at archlinux dot ca<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2009  Xyne

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.