# Copyright (C) 2009  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


package Xyne::Arch::Powerpill;
use strict;
use warnings;

use LWP::Simple qw/getstore/;
use Time::HiRes qw/time usleep/;
use File::Temp qw/tempfile/;

use Xyne::Arch::Pacman;
use Xyne::Arch::PkgD;
use Xyne::Arch::PowerpillConf;
use Xyne::Arch::Reflector;
use Xyne::Common::Functions qw/colorize divmod format_bytes format_seconds wrap_line/;
use Xyne::Common::Log;
use Xyne::Common::Metalink;

our $VERSION = '0.11';


my $arg_maps = {
  CONF => {
    '--aria2-bin'        => 'Aria2Bin',
    '--aria2-clean'      => 'Aria2Clean',
    '--aria2-silent'     => 'Aria2Silent',
    '--log'              => 'Log',
    '--no-color'         => 'NoColor',
    '--no-messages'      => 'NoMessages',
    '--pacman-bin'       => 'PacmanBin',
#    '--pacman-config' => 'PacmanConf',
    '--powerpill-config' => 'Config',
    '--reflect'          => 'Reflect',
    '--retry-limit'      => 'RetryLimit',
    '--split-size'       => 'SplitSize'
  },
  PARAMETRIC => {
    '--aria2-bin'        => 'str',
    '--log'              => 'str',
    '--pacman-bin'       => 'str',
#    '--pacman-conf'  => 'str',
    '--powerpill-config' => 'str',
    '--reflect'          => 'str',
    '--retry-limit'      => 'int',
    '--split-size'       => 'int'
  },
  POWERPILL => {
    '--aria2-bin'        => 'conf',
    '--aria2-clean'      => 'conf',
    '--aria2-silent'     => 'conf',
    '--get-metalink'     => 'action',
    '--help'             => 'action',
    '--log'              => 'conf',
    '--no-color'         => 'conf',
    '--pacman-bin'       => 'conf',
#    '--pacman-conf'  => 'conf',
#    '--pacman-help'      => 'action',
    '--powerpill-config' => 'conf',
    '--reflect'          => 'conf',
    '--retry-limit'      => 'conf',
    '--split-size'       => 'conf'
  }
};


sub new {
  my ($proto,$args) = @_;
  my $class = ref($proto) || $proto;
  my $self  = {};
  bless ($self, $class);

  $self->{Shell} = Xyne::Common::Shell->new();
  $self->{Pacman} = Xyne::Arch::Pacman->new({
    'Shell'           => $self->{Shell},
    'PacmanConf_PATH' => $args->{PacmanConf_PATH}
  });
  $self->{Pacman}->{Powerpill} = $self;
  $self->{PacmanConf} = $self->{Pacman}->{PacmanConf};
  $self->{PowerpillConf} = Xyne::Arch::PowerpillConf->new({
    'Pacman'     => $self->{Pacman},
    'PacmanConf' => $self->{PacmanConf},
    'PATH'       => $args->{PowerpillConf_PATH}
  });

  return $self;
}



sub append_to_list
{
  my ($self,$field,@pkgs) = @_;
  push @{ $self->{$field.'_LIST'} }, grep {defined($_)} @pkgs;
  $self->{$field.'_LIST_LENGTH'} += scalar @pkgs;
  my $size = $self->{$field.'_LIST_SIZE'};
  $size += $_ foreach map {$self->{Pacman}->get_sync_db_csize($_)} @pkgs;
  $self->{$field.'_LIST_SIZE'} = $size;
}

sub check_progress
{
  $|++;
  my ($self,$line) = @_;
  if (not $self->{PowerpillConf}->get_option('Aria2Silent') and $line =~ m/\S/ and $line !~ m/^\[/)
  {
    chomp $line;
    my $w = $self->{PROGRESS_LINE_WIDTH} ? $self->{PROGRESS_LINE_WIDTH} : 0;
    my $status_line = $self->{PROGRESS_LINE} ? $self->{PROGRESS_LINE} : '';
    my $line = sprintf ("%-${w}s",$line);
    $line = $self->tag_aria2_output($line);
    my $x = length($line);
    printf ("\r%-${w}s\n%-${x}s",$line,$status_line);
  }
  else
  {
    $line = '';
  }
  if (not $self->{PREVIOUS_TIME})
  {
    $self->{PREVIOUS_TIME} = time;
    return;
  }
  if (not $self->{Shell_2})
  {
    $self->{Shell_2} = Xyne::Common::Shell->new();
  }
  my $shell = $self->{Shell_2};
  my $previous_time =  $self->{PREVIOUS_TIME};
  my $current_time = time;
  my $elapsed_time = $current_time - $previous_time;
  return if $elapsed_time < 1;
  my $pacman = $self->{Pacman};
  my @pkgs = $self->get_download_queue;
  my $total_size = $self->get_download_csize;
  my $total_num = scalar @pkgs;
  my $file_to_pkg;
  if (not $self->{FILE_TO_PKG})
  {
    my $cachedir = $self->get_cachedir;
    foreach my $pkg (@pkgs)
    {
      $file_to_pkg->{$cachedir.'/'.$pacman->get_sync_db_filename($pkg)} = $pkg;
    }
    $self->{FILE_TO_PKG} = $file_to_pkg;
  }
  else
  {
    $file_to_pkg = $self->{FILE_TO_PKG};
  }
  my $completed_size = 0;
  my $completed_num = 0;
  foreach $line ($shell->execute('du -B1 '.join(' ',grep {-f} keys %{ $file_to_pkg })))
  {
    if ($line =~ m/^(\d+)\s+(.+?)\s*$/)
    {
      my ($du_size,$file) = ($1,$2);
      my $perl_size = -s $file;
      my $pkg = $file_to_pkg->{$file};
      if ($perl_size == $pacman->get_sync_db_csize($pkg))
      {
        $completed_num++;
      }
      $completed_size += ($du_size < $perl_size) ? $du_size : $perl_size;
    }
  }

  my $previous_size = $self->{PREVIOUS_SIZE} ? $self->{PREVIOUS_SIZE} : $completed_size;
  $self->{PREVIOUS_TIME} = $current_time;
  $self->{PREVIOUS_SIZE} = $completed_size;

  my $size_diff = $completed_size - $previous_size;
  my $rate = ($elapsed_time) ? $size_diff / $elapsed_time : 0;
  my $percent = ($total_size) ? int(100 * $completed_size/$total_size) : 0;
  my $remaining_size = $total_size - $completed_size;
  my $eta = ($rate) ? &format_seconds($remaining_size / $rate) : 'N/A';
  $completed_size = &format_bytes($completed_size,'bin');
  $total_size = &format_bytes($total_size,'bin');
  $rate = &format_bytes($rate,'bin');
  my $status_line = sprintf("progress: %d/%d pkgs  %s/%s (%d%%) %s/s  ETA: %s ", $completed_num, $total_num, $completed_size, $total_size, $percent, $rate, $eta);
  #$status_line = $self->tag_statement($status_line);
  my $w = $self->{PROGRESS_LINE_WIDTH} ? $self->{PROGRESS_LINE_WIDTH} : 0;
  printf "\r%-${w}s",$status_line;
  $self->{PROGRESS_LINE_WIDTH} = length($status_line);
  $self->{PROGRESS_LINE} = $status_line;
}

sub cleanup_aria2_files
{
  my ($self) = @_;
  my $cachedir = $self->get_cachedir;
  opendir(my $dh,$cachedir);
  foreach my $item (readdir($dh))
  {
    unlink $cachedir.'/'.$item if $item =~ m/\.aria2$/;
  }
  close $dh;
}

sub confirm
{
  my ($self,$question,$default) = @_;
  $question = join ("\n",map { $self->tag_question($_) } split(/\s*\n\s*/,$question));
  return $self->{Pacman}->confirm($question,$default)
}

sub confirm_download
{
  my ($self,@pkgs) = @_;
  $self->set_download(@pkgs);
  my $csize = &format_bytes($self->get_download_csize,'MiB');
  my $isize = &format_bytes($self->get_download_isize,'MiB');
  my $number = scalar @pkgs;
  my $question = $self->tag_question("The following packages will be downloaded:\n");
  $question .= $self->arrange_by_repo(@pkgs);
  $question .= $self->tag_question("Number of Packages: $number\n");
  $question .= $self->tag_question("Total Download Size: $csize\n");
  $question .= $self->tag_question("Total Installed Size: $isize\n");
  $question .= $self->tag_question("Proceed with download?");
  return $self->{Pacman}->confirm($question,'y');
}

sub arrange_by_repo
{
  my ($self,@pkgs) = @_;
  my $str = '';
  foreach my $repo ($self->{PacmanConf}->get_repos)
  {
    my @repo_pkgs;
    foreach my $pkg (@pkgs)
    {
      push @repo_pkgs, $pkg if ($pkg =~ m/^\Q$repo\E\/(.+)$/);
    }
    next if not @repo_pkgs;
    $str .= '  '.$self->maybe_colorize('[', 'bold blue').$repo.$self->maybe_colorize(']','bold blue')."\n";
    if ($self->{PacmanConf}->get_option('ShowSize'))
    {
      @repo_pkgs = $self->{Pacman}->append_csize_tags(@repo_pkgs);
    }
    $str .= &wrap_line(80, '    ', sort map {$self->{Pacman}->get_name($_)} @repo_pkgs);
  }
  return $str;
}

sub display_help
{
  my ($self) = @_;
  print <<'HELP';
POWERPILL-SPECIFIC OPTIONS
  --aria2-bin <path>         set the aria2 binary
  --aria2-clean              remove aria2 control files when done
  --aria2-silent             suppress aria2 output
  --get-metalink             print metalink
  --log <path>               set log
  --no-color                 suppress colors
  --pacman-bin <path>        set pacman bin
  --powerpill-config <path>  set powerpill configuration file
  --reflect "<args>"         use reflector
  --retry-limit <int>        set retry limit when using "--noconfirm"
  --split-size <int>         set minimum split size

See the powerpill man page for more details.
Use "powerpill -h" to see the pacman help message.
HELP
  exit;
}

sub filter_aria2_output
{
  my ($self,$output) = @_;
  return if substr($output,0,1) ne '[';
  $|++;
  chomp $output;
  my $l = $self->{ARIA2_L} ? $self->{ARIA2_L} : 0;
  printf "\r%-${l}s",$output;
  $self->{ARIA2_L} = length($output);
}

sub get_cachedir
{
  my ($self) = @_;
  foreach my $dir ($self->{PacmanConf}->get_option('CacheDir'))
  {
    return $dir if -w $dir;
  }
  return undef;
}

sub get_download_queue
{
  my ($self,$list) = @_;
  return @{ $self->{DOWNLOAD_QUEUE} };
}

sub get_download_csize
{
  my ($self,$list) = @_;
  return $self->{DOWNLOAD_CSIZE};
}

sub get_download_isize
{
  my ($self,$list) = @_;
  return $self->{DOWNLOAD_ISIZE};
}

sub get_metalink
{
  my ($self,$args,@pkgs) = @_;
  my $pacman = $self->{Pacman};
  my $metalink = Xyne::Common::Metalink -> new();
  my $split_size = $self->{PowerpillConf}->get_option('SplitSize');
  $split_size *= 1000000;
  foreach my $pkg (@pkgs)
  {
    my $file = $pacman->get_sync_db_filename($pkg);
    my $csize = $pacman->get_sync_db_csize($pkg);
    my $maxconnections = ($csize < $split_size) ? 1 : undef;
    my $metalink_args = {
      name           => $file,
      identity       => $pkg,
      version        => $pacman->get_sync_db_ver($pkg),
      size           => $csize,
      hash_type      => 'md5',
      hash           => $pacman->get_sync_db_md5sum($pkg),
      maxconnections => $maxconnections
    };
    my @urls = $pacman->get_urls($pkg);
#    my @pkgd_urls = $self->{PowerpillConf}->get_servers('PkgD');
#     if (@pkgd_urls)
#     {
#       my $url_preferences;
#       my $url_maxconnections;
#       foreach my $url (@pkgd_urls)
#       {
#         $url .= "/$file";
#         $url_preferences->{$url} = 100;
#         $url_maxconnections->{$url} = 1;
#         push @urls, $url;
#       }
#       $metalink_args->{'url_preferences'} = $url_preferences;
#       $metalink_args->{'url_maxconnections'} = $url_maxconnections;
#     };
    $metalink_args->{'urls'} = [@urls];
    $metalink_args->{'dir'} = $args->{'dir'} if defined($args->{'dir'});
    $metalink->add_file($metalink_args);
  }
  return $metalink->get_metalink;
}

sub is_action
{
  my ($self,$arg) = @_;
  return 0 if not defined($arg);
  return (defined($self->{ACTIONS}->{$arg}) and $self->{ACTIONS}->{$arg});
}

sub maybe_colorize
{
  my ($self,$str,$color) = @_;
  return &colorize($str,$color) if not $self->{PowerpillConf}->get_option('NoColor');
  return $str;
}

sub msg
{
  my ($self,$msg,$args) = @_;
  my $nocolor = $self->{PowerpillConf}->get_option('NoColor');
  my $tag = ($args->{'tag'}) ? $args->{'tag'} : '';
  if ($tag ne '' and not $nocolor and $args->{'color'})
  {
    print &colorize($tag,$args->{'color'}),$msg;
  }
  else
  {
    print $tag,$msg;
  }
  if (defined($self->{LOG}) or $self->{PowerpillConf}->get_option('Log'))
  {
    $self->{LOG} = Xyne::Common::Log->new( $self->{PowerpillConf}->get_option('Log') ) if not defined($self->{LOG});
    $self->{LOG}->log($tag.$msg);
  }
}

sub msg_error
{
  my ($self,$msg) = @_;
  $self->msg($msg,{
   'tag' => '==> ERROR ',
    'color' => 'bold red'
  });
  exit(1);
}

sub msg_normal
{
  my ($self,$msg) = @_;
  $self->msg($msg,{
   'tag' => '--> ',
    'color' => 'bold blue'
  });
}

sub msg_verified
{
  my ($self,$msg) = @_;
  $self->msg($msg,{
   'tag' => '==> VERIFIED ',
    'color' => 'bold green'
  });
}

sub msg_warning
{
  my ($self,$msg) = @_;
  $self->msg($msg,{
   'tag' => '==> WARNING ',
    'color' => 'bold red'
  });
}

sub parse_args
{
  my ($self,@args) = @_;
  my @pacman_args = ();
  my $display_pacman_help = 0;
  while (@args)
  {
    my $arg = shift @args;
    if (defined($arg_maps->{POWERPILL}->{$arg}))
    {
      if (defined($arg_maps->{PARAMETRIC}->{$arg}))
      {
        my $parameter = (@args) ? shift @args : '';
        if ($arg_maps->{PARAMETRIC}->{$arg} eq 'int')
        {
          ($parameter) = ($parameter =~ m/(\d+)/);
        }
        if (defined($arg_maps->{CONF}->{$arg}))
        {
          $self->{PowerpillConf} = $self->{PowerpillConf}->set_option_by_arg ( $arg_maps->{CONF}->{$arg} , $parameter);
        }
      }
      else
      {
        if (defined($arg_maps->{CONF}->{$arg}))
        {
          $self->{PowerpillConf} = $self->{PowerpillConf}->set_option_by_arg ( $arg_maps->{CONF}->{$arg} );
        }
        elsif ($arg eq '--help')
        {
          $self->display_help;
        }
        elsif ($arg eq '--pacman-help')
        {
          $display_pacman_help = 1;
          push @pacman_args,$arg;
        }
        elsif (defined($arg_maps->{POWERPILL}->{$arg} eq 'action'))
        {
          $self->set_action($arg);
        }
      }
    }
    else
    {
      push @pacman_args,$arg;
    }
  }
  my $operation = $self->{Pacman}->parse_args(@pacman_args);
  $self->{Pacman}->execute({},1) if $display_pacman_help;
  return $operation;
}

sub reflect
{
  my ($self) = @_;
  return if not $self->{PowerpillConf}->get_option('Reflect');
  $self->msg_normal("Retrieving mirrorlist with reflector...\n");
  my @reflector_args = split /\s+/, $self->{PowerpillConf}->get_option('Reflect');
  my $reflector = Xyne::Arch::Reflector->new();
  $reflector->parse_args(@reflector_args);
  my @servers = $reflector->get_servers;
  my @repos = $reflector->get_repos;
  my $n = scalar @servers;
  my $repos = join ' ',sort @repos;
  $self->msg_normal("Appending $n server(s) to the mirrorlists of the following repos: $repos\n");
  $self->{PacmanConf}->append_servers($_,@servers) foreach @repos;
}

sub maybe_get_pkgd_pkgs
{
  my ($self, @pkgs) = @_;
  my $cachedir = $self->{CacheDir};
  my @pkgd_urls = $self->{PowerpillConf}->get_servers('PkgD');
  if (@pkgd_urls)
  {
    @pkgs = map {$self->{Pacman}->get_sync_db_filename($_)} @pkgs;
    &Xyne::Arch::PkgD::retrieve_and_save_pkg_from_pkgd(\@pkgs, \@pkgd_urls, $cachedir);
  }
}

sub run
{
  my ($self,@args) = @_;
  $self->{PowerpillConf}->load_config;
  $self->{PacmanConf}->load_config;
  my $operation = $self->parse_args(@args);
  $self->switch_to_pacman() if $operation ne 'SYNC' or not $self->{Pacman}->is_possible_download;
  my $cachedir = $self->get_cachedir;
  if (not (defined($cachedir) or $self->is_action('--get-metalink')) )
  { $self->msg_error("no writable cache directory specified\n"); }
  $self->{CacheDir} = $cachedir;
  $self->{Pacman}->maybe_refresh_only;
  $self->{PowerpillConf}->append_servers_to_pacman_conf;
  my $download = $self->{Pacman}->get_download;
  my @pkgs = keys %{ $download };
  $self->switch_to_pacman() if not @pkgs;
  exit if not $self->confirm_download(@pkgs);
  
  $self->maybe_get_pkgd_pkgs(@pkgs) if not $self->is_action('--get-metalink');
  
  $self->set_download(@pkgs);
  $self->reflect if $self->{PowerpillConf}->get_option('Reflect');
  my $metalink = $self->get_metalink({},@pkgs);
  if ($self->is_action('--get-metalink'))
  {
    print $metalink;
    exit;
  }
  $self->run_aria2($metalink);
  my $verified = $self->verify_md5sums(@pkgs);

  if (not $verified)
  {
    my $retries = 0;
    my $RetryLimit = $self->{PowerpillConf}->get_option('RetryLimit');
    my $NoConfirm = $self->{PacmanConf}->get_option('NoConfirm');
    while (not $verified and $self->confirm('Do you want to retry the download?','y') and ($retries < $RetryLimit or not $NoConfirm))
    {
      if ($NoConfirm)
      {
        $self->msg_normal("Retrying...($retries/$RetryLimit)\n");
        $retries++;
      }
      $self->run_aria2($metalink);
      $verified = $self->verify_md5sums(@pkgs);
    }
  }
  if ($verified)
  { $self->msg_normal("All downloads have finished successfully\n"); }
  else { $self->msg_normal("Some downloads have not finished successfully.\n"); }
  $self->cleanup_aria2_files if $self->{PowerpillConf}->get_option('Aria2Clean');
  $self->switch_to_pacman();
}

sub run_aria2
{
  my ($self,$metalink) = @_;
  my $aria2_bin = $self->{PowerpillConf}->get_option('Aria2Bin');
  my $aria2_args = $self->{PowerpillConf}->get_aria2_arg_str;
  my $cachedir =$self->get_cachedir;
  $aria2_args =~ s/%CacheDir%/$cachedir/g;
  my $cmd = "$aria2_bin $aria2_args --metalink-file=- <<'METALINK'\n$metalink\nMETALINK\n";
#   # Workaround when aria2 briefly broke pipe support.
#   if (1)
#   {
#     print "piping\n";
#     $cmd = "$aria2_bin $aria2_args --metalink-file=- <<'METALINK'\n$metalink\nMETALINK\n";
#   }
#   # use a temp file now that aria2c no longer accepts metalink pipes
#   else
#   {
#     my ($fh, $fname) = &tempfile('/tmp/powerpill-XXXX', SUFFIX => '.metalink', UNLINK => 1);
#     print $fh $metalink;
#     close $fh;
#     $cmd = "$aria2_bin $aria2_args --metalink-file='$fname'\n";
#   }
  $self->{Shell}->set_callback(sub {$self->check_progress(@_)});
  $self->{Shell}->execute($cmd);
  $self->{Shell}->set_callback(undef);
  print "\r\n";
}

sub set_action
{
  my ($self,$arg) = @_;
  return if not defined($arg);
  if ($arg_maps->{'POWERPILL'}->{$arg} eq 'action')
  {
    $self->{ACTIONS}->{$arg} = 1;
    return 1;
  }
}

sub set_download
{
  my ($self,@pkgs) = @_;
  $self->{DOWNLOAD_QUEUE} = [@pkgs];
  my $size = 0;
  $size += $_ foreach map {$self->{Pacman}->get_sync_db_csize($_)} @pkgs;
  $self->{DOWNLOAD_CSIZE} = $size;
  $size = 0;
  $size += $_ foreach map {$self->{Pacman}->get_sync_db_isize($_)} @pkgs;
  $self->{DOWNLOAD_ISIZE} = $size;
}

sub switch_to_pacman
{
  my ($self,@msgs) = @_;
  $self->msg_normal($_) foreach @msgs;

  my $pacman = $self->{Pacman};
  if (not (
    $self->is_action('--get-metalink') or
    $pacman->{OPERATION} eq 'SYNC' and $pacman->{ARGS}->{'--downloadonly'}
  ))
  {
    $pacman->execute({},1);
  }
  exit;
}

sub tag_aria2_output
{
  my ($self,$str) = @_;
  return $self->maybe_colorize('aria2: ','cyan').$str;
}

sub tag_question
{
  my ($self,$question) = @_;
  return $self->maybe_colorize(':: ','bold blue').$question;
}

sub tag_statement
{
  my ($self,$str) = @_;
  return $self->maybe_colorize('--> ','bold blue').$str;
}

sub verify_md5sums
{
  my ($self,@pkgs) = @_;
  my $cachedir = $self->get_cachedir;
  my %file_to_pkg;
  @file_to_pkg{map {$cachedir.'/'.$self->{Pacman}->get_sync_db_filename($_)} @pkgs} = @pkgs;
  my $passed = 1;
  foreach my $line ($self->{Pacman}->get_disk_md5sums(keys %file_to_pkg))
  {
    if ($line =~ m/^([0-9A-Fa-f]{32})\s+([^\s]+)$/)
    {
      my ($file,$md5sum) = ($2,$1);
      if ($md5sum eq $self->{Pacman}->get_sync_db_md5sum($file_to_pkg{$file}))
      {
        $self->msg_verified($file."\n");
      }
      else
      {
        $passed = 0;
        $self->msg_warning("$file: checksum failed\n");
        if (-e $file.'.aria2')
        {
          my $question = "A matching aria2 control file was found.\n";
          $question .= "It may be possible to resume this download.\n";
          $question .= "Do you want to delete the files anyway?";
          if ($self->confirm($question,'y'))
          {
            unlink $file;
            unlink $file.'.aria2';
          }
        }
        else
        {
          unlink $file if $self->confirm("File $file is corrupted. Do you want to delete it?",'y');
        }
      }
    }
  }
  return $passed;
}

1;

__END__

=head1 NAME

Xyne::Arch::Powerpill - A pacman wrapper and download-accelerator class.

=head1 SYNOPSIS

  use Xyne::Arch::Powerpill;

=head1 DESCRIPTION

Powerpill is intended to be a fully-functional wrapper for pacman
that uses aria2 to accelerate downloads while leaving all other
operations to pacman itself.

=head2 EXPORT

None by default.

=head1 AUTHOR

Xyne <lt>xyne at archlinux dot ca<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2009  Xyne

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.