# Copyright (C) 2009  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


package Xyne::Arch::PowerpillConf;
use strict;
use warnings;

use Xyne::Common::Functions qw/strip_trailing_slash/;

our $VERSION = '0.04';

my $OPTIONS = {
  'Aria2Bin'    => 'str',
  'Aria2Silent' => 'switch',
  'Aria2Clean'  => 'switch',
  'Log'         => 'str',
  'NoColor'     => 'switch',
#  'NoMessages'  => 'switch',
  'PacmanBin'   => 'str',
#  'PacmanConf'  => 'str',
  'Reflect'     => 'str',
  'RetryLimit'  => 'int',
  'SplitSize'   => 'int'
};

my $DEFAULTS = {
  'Aria2Bin'   => '/usr/bin/aria2c',
  'PacmanBin'  => '/usr/bin/pacman',
  'SplitSize'  => '5',
  'RetryLimit' => '3'
};

my $ARIA2_DEFAULTS = {
  'allow-overwrite'                 => 'true',
  'auto-file-renaming'              => 'false',
  'continue'                        => '',
  'file-allocation'                 => 'none',
  'ftp-pasv'                        => 'true',
  'max-concurrent-downloads'        => '45',
  'max-tries'                       => '2',
  'metalink-enable-unique-protocol' => 'false',
  'metalink-servers'                => '45',
# removed in aria2 version 1.6  
#  'retry-wait'                      => '2',
  'split'                           => '10',
  'summary-interval'                => '0',
  'timeout'                         => '5',
};

my $ARIA2_MANDATORY_ARGS = {
# aria2 version 1.6 broke metalink piping
#  'metalink-file'   => '-',
  'dir'             => '%CacheDir%',
  'file-allocation' => 'none'
};

my @NONREPOS = qw/options aria2_options/;
my @DEFAULT_PATHS = (
  $ENV{'HOME'}.'/.config/powerpill/powerpill.conf',
  '/etc/powerpill.conf'
);


sub new {
  my ($proto,$args) = @_;
  my $class = ref($proto) || $proto;
  my $self  = {};
  bless ($self, $class);

  if (defined($args->{PATH}) and -e $args->{PATH})
  {
    $self->{PATH} = $args->{PATH};
  }
  else
  {
    foreach my $path (@DEFAULT_PATHS)
    {
      if (-e $path)
      {
        $self->{PATH} = $path;
        last;
      }
    }
  }
  
  $self->{Pacman} = $args->{Pacman};
  $self->{PacmanConf} = $args->{PacmanConf};
  
  $self->{options} = {};
  $self->{aria2_options} = {};
  
  return $self;
}





sub append_servers
{
  my ($self,$repo,@servers) = @_;
  return if not defined($repo) or grep {$_ eq $repo} @NONREPOS;
  foreach my $server (@servers)
  {
#    print "appending $server to $repo\n";
    $server = &strip_trailing_slash($server);
    $server =~ s/\$repo(?!\w)/$repo/g;
    push @{ $self->{repos}->{$repo} }, $server if not grep {$_ eq $server} @{ $self->{repos}->{$repo} };
  }
}

sub append_servers_to_pacman_conf
{
  my ($self) = @_;
  foreach my $repo ($self->{PacmanConf}->get_repos)
  {
    $self->{PacmanConf}->append_servers($repo,$self->get_servers($repo));
  }
}

sub get_aria2_arg_str
{
  my ($self) = @_;
  my %options = (
    %{ $ARIA2_DEFAULTS },
    %{ $self->{aria2_options} },
    %{ $ARIA2_MANDATORY_ARGS }
  );
  my $str;
  foreach my $option (keys %options)
  {
    my $value = $options{$option};
    if ($value =~ m/\S/)
    {
      $str .= " --$option=$options{$option}";
    }
    else
    {
      $str .= " --$option";
    }
  }
  return $str;
}

sub get_option
{
  my ($self,$option) = @_;
  return if not defined($OPTIONS->{$option});
  if (defined($self->{options}->{$option}))
  {
    return $self->{options}->{$option};
  }
  elsif (defined($DEFAULTS->{$option}))
  {
    return $DEFAULTS->{$option};
  }
  elsif ($OPTIONS->{$option} eq 'switch')
  {
    return 0;
  }
  return;
}

sub get_option_set_by_arg
{
  my ($self,$option) = @_;
  return $self->get_option($option) if defined($self->{args}->{$option});
}

sub get_servers
{
  my ($self,$repo) = @_;
  return () if not defined($repo) or not defined($self->{repos}->{$repo});
  return grep {$_} @{ $self->{repos}->{$repo} };
}


sub load_config
{
  my ($self,$args,@loaded) = @_;
  my $file_path = defined($args->{'file_path'}) ? $args->{'file_path'} : $self->{PATH};
  # Return if this file has been loaded already. This avoids cyclic recursion.
  return $self if (grep {$_ eq $file_path} @loaded);
  my $repo = defined($args->{'repo'}) ? $args->{'repo'} : '';

  return if not $file_path or not -e $file_path;
  open (my $fh, '<', $file_path) or die "error: failed to open $file_path ($!)";
  while ( defined(my $line = <$fh>) )
  {
    next if ($line =~ m/^\s*#/);
    if ($line =~ m/^\s*\[([^\s\]]+)\]\s*$/)
    {
      $repo = $1;
      push @{ $self->{repo_list} }, $repo if not grep {$_ eq $repo} @NONREPOS;
      $args->{'repo'} = $repo;
    }
    elsif ($line =~ m/^\s*(\S+)\s*=\s*(.+)\s*$/)
    {
      if ($repo eq 'options')
      {
        $self->set_option($1,$2);
      }
      elsif ($repo eq 'aria2_options')
      {
        $self->set_aria2_option($1,$2);
      }
      elsif ($1 eq 'Include')
      {
        $args->{'file_path'} = $2;
        $self->load_config($args,@loaded,$file_path);
      }
      elsif ($1 eq 'Server' and $repo)
      {
        $self->append_servers($repo,$2);
      }
    }

    elsif ($line =~ m/^\s*(\S+)\s*$/ and $repo eq 'options')
    {
      $self->set_option($1);
    }
  }
  close $fh;
  return $self;
}

sub load_config_by_arg
{
  my ($self,$path) = @_;
  my $config = $self->new({
    PATH       => $path,
    Pacman     => $self->{Pacman},
    PacmanConf => $self->{PacmanConf}
  });
  $config->load_config;
  foreach my $option (keys %{ $self->{args} })
  {
    $config->set_option_by_arg($option, $self->get_option_set_by_arg($option));
  }
  $self = $config;
  return $self;
}

sub set_aria2_option
{
  my ($self,$option,$value) = @_;
  $self->{aria2_options}->{$option} = $value;
}

sub set_option
{
  my ($self,$option,$value) = @_;
  return 0 if not defined($OPTIONS->{$option});
  if ($option eq 'PacmanBin')
  {
    $self->{Pacman}->set_binary($value);
    return 1;
  }
#  if ($option eq 'PacmanConf')
#  {
#    $self->{PacmanConf}->set_path($value);
#  }
  my $type = $OPTIONS->{$option};
  if ($type eq 'switch')
  {
    if (defined($value) and not $value)
    {
      delete($self->{options}->{$option});
    }
    else
    {
      $self->{options}->{$option} = 1;
    }
  }
  elsif ($type eq 'int')
  {
    $self->{options}->{$option} = $value if $value =~ m/^\d+$/;
  }
  else
  {
    $self->{options}->{$option} = $value if defined($value);
  }
  return 1;
}

sub set_option_by_arg
{
  my ($self,$option,$value) = @_;
  if ($option eq 'Config')
  {
    return $self->load_config_by_arg($value);
  }
  $self->{args}->{$option} = 1 if $self->set_option($option,$value);
  return $self;
}

1;

__END__

=head1 NAME

Xyne::Arch::PowerpillConf - Powerpill configuration file class.

=head1 SYNOPSIS

  use Xyne::Arch::PowerpillConf;

=head1 DESCRIPTION

A class to load and parse powerpill configuration files and
store powerpill settings, including arguments passed to it from
powerpill objects.

=head2 EXPORT

None by default.

=head1 AUTHOR

Xyne <lt>xyne at archlinux dot ca<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2009  Xyne

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

=cut
